/* packet-rtpmidi.c
 *
 * Routines for RFC 4695 RTP-MIDI dissection
 * Copyright 2006, Tobias Erichsen <t.erichsen@gmx.de>
 *
 * $Id: packet-rtpmidi.c 18196 2006-05-21 04:49:01Z sahlberg $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 * RTP-MIDI ( RFC-4695) is a payload type for transmitting MIDI-data over RTP-
 * packets.  This payload type can be used on an RTP-conversation that has been
 * established either by means of a session establishment-protocol like SIP or
 * something more lightweighted like the Apple network-midi session establishment.
 * Using RTP-Midi payload is also possible without session-establishment.
 *
 * Here are some links:
 * 
 * http://www.cs.berkeley.edu/~lazzaro/rtpmidi/
 * http://www.faqs.org/rfcs/rfc4695.html (soon to be published)
 * http://www.midi.org/

 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <epan/packet.h>
#include <epan/prefs.h>

#include <stdio.h>
#include <string.h>
#include "packet-rtpmidi.h"
#include <epan/conversation.h>
#include <epan/tap.h>

/*  rtp_midi_payload_type_value is the value used globally
	to set the appropriate payload type
    saved_pt_value is a temporary place to save the value
    	so we can properly reinitialize when the settings
    	get changed
*/
static guint rtp_midi_payload_type_value = 97;
static guint saved_payload_type_value;


/* RTP Event Fields */

static int proto_rtp_midi					= -1;
static int hf_rtp_midi_bflag				= -1;
static int hf_rtp_midi_jflag				= -1;
static int hf_rtp_midi_zflag				= -1;
static int hf_rtp_midi_pflag				= -1;
static int hf_rtp_midi_shortlen				= -1;
static int hf_rtp_midi_longlen				= -1;

static int hf_rtp_midi_sflag				= -1;
static int hf_rtp_midi_yflag				= -1;
static int hf_rtp_midi_aflag				= -1;
static int hf_rtp_midi_hflag				= -1;
static int hf_rtp_midi_totchan				= -1;
static int hf_rtp_midi_check_seq_num		= -1;

static int hf_rtp_midi_deltatime1			= -1;
static int hf_rtp_midi_deltatime2			= -1;
static int hf_rtp_midi_deltatime3			= -1;
static int hf_rtp_midi_deltatime4			= -1;

static int hf_rtp_midi_status				= -1;
static int hf_rtp_midi_rsstatus				= -1;
static int hf_rtp_midi_command				= -1;
static int hf_rtp_midi_rscommand			= -1;
static int hf_rtp_midi_channel				= -1;
static int hf_rtp_midi_rschannel			= -1;
static int hf_rtp_midi_note					= -1;
static int hf_rtp_midi_velocity				= -1;
static int hf_rtp_midi_pressure				= -1;
static int hf_rtp_midi_controller			= -1;
static int hf_rtp_midi_controller_value		= -1;
static int hf_rtp_midi_program				= -1;
static int hf_rtp_midi_channel_pressure		= -1;

static int hf_rtp_midi_manu_short			= -1;
static int hf_rtp_midi_manu_long			= -1;

static int hf_rtp_midi_sysjour_toc_s		= -1;
static int hf_rtp_midi_sysjour_toc_d		= -1;
static int hf_rtp_midi_sysjour_toc_v		= -1;
static int hf_rtp_midi_sysjour_toc_q		= -1;
static int hf_rtp_midi_sysjour_toc_f		= -1;
static int hf_rtp_midi_sysjour_toc_x		= -1;
static int hf_rtp_midi_sysjour_len			= -1;

static int hf_rtp_midi_chanjour_sflag		= -1;
static int hf_rtp_midi_chanjour_chan		= -1;
static int hf_rtp_midi_chanjour_hflag		= -1;
static int hf_rtp_midi_chanjour_len			= -1;
static int hf_rtp_midi_chanjour_toc_p		= -1;
static int hf_rtp_midi_chanjour_toc_c		= -1;
static int hf_rtp_midi_chanjour_toc_m		= -1;
static int hf_rtp_midi_chanjour_toc_w		= -1;
static int hf_rtp_midi_chanjour_toc_n		= -1;
static int hf_rtp_midi_chanjour_toc_e		= -1;
static int hf_rtp_midi_chanjour_toc_t		= -1;
static int hf_rtp_midi_chanjour_toc_a		= -1;

static int hf_rtp_midi_sysex_chan			= -1;






/* RTP MIDI fields defining a subtree */

static gint ett_rtp_midi           = -1;
static gint ett_rtp_midi_commands  = -1;
static gint ett_rtp_midi_journal   = -1;
static gint ett_rtp_midi_command   = -1;
static gint ett_rtp_midi_systemjournal = -1;
static gint ett_rtp_midi_channeljournals = -1;
static gint ett_rtp_midi_systemchapters = -1;
static gint ett_rtp_midi_channeljournal =-1;
static gint ett_rtp_midi_channelchapters = -1;
static gint ett_rtp_midi_sysex_data = -1;

void proto_reg_handoff_rtp_midi(void);



static unsigned int
decodetime(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int cmd_offset, unsigned int cmd_len)
{
	guint8			octet;
	unsigned int	consumed;
	unsigned int    i;
	unsigned long   deltatime;

	consumed=0;
	deltatime=0;


	for (i=0;i<4;i++) {

		if ((cmd_offset+consumed)>=cmd_len) {
			return 0;
		}

		octet = tvb_get_guint8(tvb, offset+cmd_offset+consumed );
		deltatime=(deltatime << 7) | (octet&0x7f);
		consumed++;

		if ((octet&0x80)!=0x80) {
			break;
		}
	}

	switch (consumed) {
		case 1: 
		    proto_tree_add_uint ( tree, hf_rtp_midi_deltatime1, tvb, offset+cmd_offset, consumed, deltatime);
			break;
		case 2: 
		    proto_tree_add_uint ( tree, hf_rtp_midi_deltatime2, tvb, offset+cmd_offset, consumed, deltatime);
			break;
		case 3: 
		    proto_tree_add_uint ( tree, hf_rtp_midi_deltatime3, tvb, offset+cmd_offset, consumed, deltatime);
			break;
		case 4: 
		    proto_tree_add_uint ( tree, hf_rtp_midi_deltatime4, tvb, offset+cmd_offset, consumed, deltatime);
			break;

	}

	return consumed;
}

static unsigned int
decodemidi(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree, unsigned int offset, unsigned int cmd_offset, unsigned int cmd_len, guint8 *runningstatus, unsigned int *rsoffset, char *text)
{
	guint8			octet;
	guint8			octet2;
	guint8			octet3;
	guint8			cmd;
	guint8			channel;
	unsigned int	consumed;
	proto_item		*command_item     = NULL;
	proto_tree		*command_tree     = NULL;
	proto_tree		*sysexdata_tree   = NULL;
	gboolean		foundend;
	gboolean		usingrs;
    guint16         wordvalue;
	char			helptext[200];
	char			entry[200];
	char			segtype[200];
	guint8			manu1;
	guint16			manu2;
	guint8			subid1;
	guint8			subid2;
	guint8			sysexchan;
	unsigned int	helpoffset;

	consumed=0;
	usingrs=FALSE;
	strcpy(text,"");

	if (cmd_offset>=cmd_len) {
		return 0;
	}

	octet = tvb_get_guint8(tvb, offset+cmd_offset);

	/* see if we have a running status message*/
	if ((octet & 0x80)==0) {
		/* if we have no running status yet -> error*/
		if (((*runningstatus) & 0x80)==0) {
			return 0;
		}
		octet=*runningstatus;
		usingrs=TRUE;
	} else {
		/* we only have consumed anything yet, if we didnt have running status*/
		consumed++;
		/* we need to store the location for the runningstatus to reference the correct octet in decoded-listing*/
		*rsoffset=offset+cmd_offset;
	}


	
	if (octet>=0xf8) {						/* midi realtime-data -> one octet*/
		switch (octet) {
			case 0xf8:
				strcpy(text,"Realtime: Midi Clock");
				break;
			case 0xf9:
				strcpy(text,"Realtime: Midi Tick");
				break;
			case 0xfa:
				strcpy(text,"Realtime: Start song");
				break;
			case 0xfb:
				strcpy(text,"Realtime: Continue song");
				break;
			case 0xfc:
				strcpy(text,"Realtime: Stop song");
				break;
			case 0xfd:
				strcpy(text,"Realtime: Undefined");
				break;
			case 0xfe:
				strcpy(text,"Realtime: Active sensing");
				break;
			case 0xff:
				strcpy(text,"Realtime: System reset");
				break;
		}
		command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, 1, text);
		command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
	    proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
	} else if (octet>=0xf0) {				/* common-data*/
		*runningstatus=0;
		switch(octet) {
			case 0xf0:
				foundend=FALSE;
				while((cmd_offset+consumed)<cmd_len) {
					octet2 = tvb_get_guint8(tvb, offset+cmd_offset+consumed );
					if ((octet2 & 0x80)==0x80) {
						foundend=TRUE;
						switch (octet2) {
							case 0xf0:
							case 0xf7:
							case 0xf5:
								consumed++;
								if (octet2==0xf0) {
									strcpy(segtype,"Common: System-Exclusive (first segment)");
								} else if (octet2==0xf7){
									strcpy(segtype,"Common: System-Exclusive (complete segment)");
								} else {
									strcpy(segtype,"Common: System-Exclusive (complete segment, non 0xf7-status termination)");
								}
								strcpy(text,segtype);
								command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
								command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
								command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-2, "Sysex-data");
								if (consumed>2) {
									manu1 = tvb_get_guint8(tvb, offset+cmd_offset+1);
									switch (manu1) {
										case 0:
											if (consumed>4) {
												manu2=tvb_get_ntohs(tvb,offset+cmd_offset+2);
											    sysexdata_tree = proto_item_add_subtree( command_item, ett_rtp_midi_sysex_data);
												proto_tree_add_uint (sysexdata_tree, hf_rtp_midi_manu_short, tvb, offset+cmd_offset+1, 1, manu1);
												proto_tree_add_uint (sysexdata_tree, hf_rtp_midi_manu_long, tvb, offset+cmd_offset+2, 2, manu2);
											    sprintf(text,"%s [manu=%s]",segtype,val_to_str( manu2, rtp_midi_manu_long_values, "Unknown: 0x%02x"));
											}
											break;
										case 0x7e: /* non realtime*/
											if (consumed>4) {
												sysexchan = tvb_get_guint8(tvb, offset+cmd_offset+2 );
												subid1=tvb_get_guint8(tvb,offset+cmd_offset+3);
												switch (subid1) {
													case 0x01: /* Sample Dump Header*/
														sprintf(entry,"Universal NRT: Sample Dump Header");
														sprintf(text,"%s [Universal NRT: Sample Dump Header]",segtype);
														break;
													case 0x02: /* Sample Dump Data Packet*/
														sprintf(entry,"Universal NRT: Sample Dump Data Packet");
														sprintf(text,"%s [Universal NRT: Sample Dump Data Packet]",segtype);
														break;
													case 0x03: /* Sample Dump Request*/
														sprintf(entry,"Universal NRT: Sample Dump Request");
														sprintf(text,"%s [Universal NRT: Sample Dump Request]",segtype);
														break;
													case 0x04: /* MTC Cueing*/
														sprintf(entry,"Universal NRT: MTC cueing");
														sprintf(text,"%s [Universal NRT: MTC cueing]",segtype);
														break;
													case 0x05: /* Loop Points*/
														sprintf(entry,"Universal NRT: Loop Points");
														sprintf(text,"%s [Universal NRT: Loop Points]",segtype);
														break;
													case 0x06: /* ID request*/
														sprintf(entry,"Universal NRT: General Information");
														sprintf(text,"%s [Universal NRT: General Information]",segtype);
														sysexdata_tree = proto_item_add_subtree( command_item, ett_rtp_midi_sysex_data);
														proto_tree_add_uint ( sysexdata_tree, hf_rtp_midi_sysex_chan, tvb, offset+cmd_offset+2, 1, sysexchan);
														command_item = proto_tree_add_text(sysexdata_tree, tvb, offset+cmd_offset+3, 1, entry);
														if (consumed>5) {
															
															subid2=tvb_get_guint8(tvb,offset+cmd_offset+4);
															switch (subid2) {

																case 1:
																	sprintf(entry,"Universal NRT: General Information: Identify Request");
																	sprintf(text,"%s [Universal NRT: General Information: Identify Request]",segtype);
																	command_item = proto_tree_add_text(sysexdata_tree, tvb, offset+cmd_offset+4, 1, entry);
																	break;

																case 2:
																	sprintf(entry,"Universal NRT: General Information: Identify Response");
																	sprintf(text,"%s [Universal NRT: General Information: Identify Response]",segtype);
																	command_item = proto_tree_add_text(sysexdata_tree, tvb, offset+cmd_offset+4, 1, entry);
																	if (consumed>6) {
																		manu1=tvb_get_guint8(tvb,offset+cmd_offset+5);
																		if (manu1!=0) {
																			helpoffset=6;
																			sprintf(entry,"Manufacturer: %s",val_to_str( manu1, rtp_midi_manu_short_values, "Unknown: 0x%02x"));
																			proto_tree_add_uint (sysexdata_tree, hf_rtp_midi_manu_short, tvb, offset+cmd_offset+5, 1, manu1);
																		} else if (consumed>8) {
																			helpoffset=8;
																			manu2=tvb_get_ntohs(tvb,offset+cmd_offset+6);
																			sprintf(entry,"Manufacturer: %s",val_to_str( manu2, rtp_midi_manu_short_values, "Unknown: 0x%02x"));
																			proto_tree_add_uint (sysexdata_tree, hf_rtp_midi_manu_short, tvb, offset+cmd_offset+5, 1, manu1);
																			proto_tree_add_uint (sysexdata_tree, hf_rtp_midi_manu_long, tvb, offset+cmd_offset+6, 2, manu2);
																		}
																	}
																	break;

																default:
																	sprintf(entry,"Universal NRT: General Information: Unknown (0x%02x)",subid2);
																	sprintf(text,"%s [Universal NRT: General Information: Unknown(0x%02x)]",segtype,subid2);
																	command_item = proto_tree_add_text(sysexdata_tree, tvb, offset+cmd_offset+4, 1, entry);
														
															}
														} else {
														}
														return consumed;
														break;
													case 0x07: /* File dump*/
														sprintf(entry,"Universal NRT: File dump");
														sprintf(text,"%s [Universal NRT: File dump]",segtype);
														break;
													case 0x08: /* MIDI Tuning Standard*/
														sprintf(entry,"Universal: MIDI Tuning standard");
														sprintf(text,"%s [Universal NRT: MIDI Tuning standard]",segtype);
														break;
													case 0x09: /* GM */
														sprintf(entry,"Universal NRT: General Midi");
														sprintf(text,"%s [Universal NRT: General Midi]",segtype);
														break;
													case 0x0a: /* DLS */
														sprintf(entry,"Universal NRT: DLS");
														sprintf(text,"%s [Universal NRT: DLS]",segtype);
														break;
													case 0x7b: /* End of file */
														sprintf(entry,"Universal NRT: End of File");
														sprintf(text,"%s [Universal NRT: End of File]",segtype);
														break;
													case 0x7c: /* Wait */
														sprintf(entry,"Universal NRT: Wait");
														sprintf(text,"%s [Universal NRT: Wait]",segtype);
														break;
													case 0x7d: /* Cancel */
														sprintf(entry,"Universal NRT: Cancel");
														sprintf(text,"%s [Universal NRT: Cancel]",segtype);
														break;
													case 0x7e: /* NAK */
														sprintf(entry,"Universal NRT: NAK");
														sprintf(text,"%s [Universal NRT: NAK]",segtype);
														break;
													case 0x7f: /* ACK */
														sprintf(entry,"Universal NRT: ACK");
														sprintf(text,"%s [Universal NRT: ACK]",segtype);
														break;
													default:
														sprintf(entry,"Universal NRT: unknown (0x%02x)",subid1);
														sprintf(text,"%s [Universal NRT: unknown (0x%02x)]",segtype,subid1);
												}
												
												sysexdata_tree = proto_item_add_subtree( command_item, ett_rtp_midi_sysex_data);
												proto_tree_add_uint ( sysexdata_tree, hf_rtp_midi_sysex_chan, tvb, offset+cmd_offset+2, 1, sysexchan);
												command_item = proto_tree_add_text(sysexdata_tree, tvb, offset+cmd_offset+3, 1, entry);
											}
											break;
										case 0x7f: /* realtime */
											if (consumed>4) {
												sysexchan = tvb_get_guint8(tvb, offset+cmd_offset+2 );
												subid1=tvb_get_guint8(tvb,offset+cmd_offset+3);
												switch (subid1) {
													case 0x01: /* Time Code */
														sprintf(entry,"Universal RT: MIDI Time Code");
														sprintf(text,"%s [Universal RT: MIDI Time Code]",segtype);
														break;
													case 0x03: /* Notation Information */
														sprintf(entry,"Universal RT: Notation Information");
														sprintf(text,"%s [Universal RT: Notation Information]",segtype);
														break;
													case 0x04: /* Device Control */
														sprintf(entry,"Universal RT: Device Control");
														sprintf(text,"%s [Universal RT: Device Control]",segtype);
														break;
													case 0x05: /* MTC Cueing */
														sprintf(entry,"Universal RT: MTC Cueing");
														sprintf(text,"%s [Universal RT: MTC Cueing]",segtype);
														break;
													case 0x08: /* MIDI Tunging standard */
														sprintf(entry,"Universal RT: MIDI Tuning standard");
														sprintf(text,"%s [Universal RT: MIDI Tuning standard]",segtype);
														break;
													default:
														sprintf(entry,"Universal RT: unknown (0x%02x)",subid1);
														sprintf(text,"%s [Universal RT: unknown (0x%02x)]",segtype,subid1);
												}
												sysexdata_tree = proto_item_add_subtree( command_item, ett_rtp_midi_sysex_data);
												proto_tree_add_uint ( sysexdata_tree, hf_rtp_midi_sysex_chan, tvb, offset+cmd_offset+2, 1, sysexchan);
												command_item = proto_tree_add_text(sysexdata_tree, tvb, offset+cmd_offset+3, 1, entry);
											}
											break;
										default:
											sysexdata_tree = proto_item_add_subtree( command_item, ett_rtp_midi_sysex_data);
											proto_tree_add_uint (sysexdata_tree, hf_rtp_midi_manu_short, tvb, offset+cmd_offset+1, 1, manu1);
											sprintf(text,"Common: System-Exclusive (complete segment) [manu=%s]",val_to_str( manu1, rtp_midi_manu_short_values, "Unknown: 0x%02x"));
									}

								}
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset+consumed-1, 1, octet2);
								return consumed;
								break;
							case 0xf4:
								consumed++;
								strcpy(text,"Common: System-Exclusive (cancel)");
								command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
								command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
								command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-2, "Sysex-data");
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset+consumed-1, 1, octet2);
								return consumed;
								break;
						}
						break;
					}
					consumed++;
				}
				/* if we didn't find proper segment-terminator -> error */
				if (TRUE==foundend) {
				  strcpy(text,"Common: System-Exclusive (invalid termination)");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "Sysex-data");
				  return consumed;
				} else {
				  strcpy(text,"Common: System-Exclusive (no termination)");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "Sysex-data");
				  return 0;
				}

				break;
			case 0xf1:
				foundend=FALSE;
				while((cmd_offset+consumed)<cmd_len) {
					octet2 = tvb_get_guint8(tvb, offset+cmd_offset+consumed );
					if ((octet2 & 0x80)==0x80) {
						foundend=TRUE;
						break;
					}
					consumed++;
				}
				/* if we didn't find proper segment-terminator -> error */
				if (TRUE==foundend) {
				  strcpy(text,"Common: Undefined");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "data");
				  return consumed;
				} else {
				  strcpy(text,"Common: Undefined (no termination)");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "data");
				  return 0;
				}
				break;
			case 0xf2:
				if ((cmd_offset+consumed+2)>cmd_len) {
					strcpy(text,"Common: Song position");
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
					return 0;
				}
				octet2=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				octet3=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				wordvalue=octet3 & 0x7f;
				wordvalue<<=7;
				wordvalue|=octet2 & 0x7f;
				sprintf(text,"Common: Song position (0x%04x)",wordvalue);
				command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				sprintf(helptext,"Position pointer: 0x%02x",wordvalue);
				command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, 2, helptext);
				break;
			case 0xf3:
				if ((cmd_offset+consumed+1)>cmd_len) {
					strcpy(text,"Common: Song select");
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
					return 0;
				}
				octet2=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				sprintf(text,"Common: Song select (0x%04x)",octet2 &0x7f);
				command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				sprintf(helptext,"Song: 0x%02x",octet2);
				command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, 1, helptext);
				break;
			case 0xf4:
				foundend=FALSE;
				while((cmd_offset+consumed)<cmd_len) {
					octet2 = tvb_get_guint8(tvb, offset+cmd_offset+consumed );
					if ((octet2 & 0x80)==0x80) {
						foundend=TRUE;
						break;
					}
					consumed++;
				}
				/* if we didn't find proper segment-terminator -> error */
				if (TRUE==foundend) {
				  strcpy(text,"Common: Undefined");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "data");
				  return consumed;
				} else {
				  strcpy(text,"Common: Undefined (no termination)");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "data");
				  return 0;
				}
				break;
			case 0xf5:
				foundend=FALSE;
				while((cmd_offset+consumed)<cmd_len) {
					octet2 = tvb_get_guint8(tvb, offset+cmd_offset+consumed );
					if ((octet2 & 0x80)==0x80) {
						foundend=TRUE;
						break;
					}
					consumed++;
				}
				/* if we didn't find proper segment-terminator -> error */
				if (TRUE==foundend) {
				  strcpy(text,"Common: Undefined");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "data");
				  return consumed;
				} else {
				  strcpy(text,"Common: Undefined (no termination)");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "data");
				  return 0;
				}
				break;
			case 0xf6:
				  strcpy(text,"Common: Tune request");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				break;
			case 0xf7:
				foundend=FALSE;
				while((cmd_offset+consumed)<cmd_len) {
					octet2 = tvb_get_guint8(tvb, offset+cmd_offset+consumed );
					if ((octet2 & 0x80)==0x80) {
						foundend=TRUE;
						switch (octet2) {
							case 0xf0:
								consumed++;
								strcpy(text,"Common: System-Exclusive (middle segment)");
								command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
								command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
								command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-2, "Sysex-data");
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset+consumed-1, 1, octet2);
								return consumed;
								break;
							case 0xf5:
							case 0xf7:
								consumed++;
								if (octet2==0xf7) {
									strcpy(text,"Common: System-Exclusive (last segment)");
								} else {
									strcpy(text,"Common: System-Exclusive (last segment, non 0xf7-status termination)");
								}
								command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
								command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
								command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-2, "Sysex-data");
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset+consumed-1, 1, octet2);
								return consumed;
								break;
							case 0xf4:
								consumed++;
								strcpy(text,"Common: System-Exclusive (cancel)");
								command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
								command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
								command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-2, "Sysex-data");
								proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset+consumed-1, 1, octet2);
								return consumed;
								break;
						}
						break;
					}
					consumed++;
				}
				/* if we didn't find proper segment-terminator -> error */
				if (TRUE==foundend) {
				  strcpy(text,"Common: System-Exclusive (invalid termination)");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "Sysex-data");
				  return consumed;
				} else {
				  strcpy(text,"Common: System-Exclusive (no termination)");
				  command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
				  command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
				  proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
				  command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, consumed-1, "Sysex-data");
				  return 0;
				}

				break;
		}
		/* common-data resets runningstatus */
	} else if (octet>=0x80) {				/* channel data */
		cmd=octet & 0xf0;
		channel = octet & 0x0f;
		/* set running status to current status-octet */
		*runningstatus=octet;
		switch(cmd) {
			case 0x80:
				if ((cmd_offset+consumed+2)>cmd_len) {
					sprintf(text,"Voice: Note off (c=%d)",channel+1);
					if (usingrs) {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-(cmd_offset+1), text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_rsstatus, tvb, *rsoffset, 1, octet);
					} else {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-cmd_offset, text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
					}
					return 0;
				}
				octet2=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				octet3=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				sprintf(text,"Voice: Note off (c=%d,n=%d,v=%d)",channel+1,octet2&0x7f,octet3&0x7f);
				if (usingrs) {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rscommand, tvb, *rsoffset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rschannel, tvb, *rsoffset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_note, tvb, offset+cmd_offset, 1, octet2 & 0x7f);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_velocity, tvb, offset+cmd_offset+1, 1, octet3 & 0x7f);
				} else {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_command, tvb, offset+cmd_offset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_channel, tvb, offset+cmd_offset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_note, tvb, offset+cmd_offset+1, 1, octet2 & 0x7f);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_velocity, tvb, offset+cmd_offset+2, 1, octet3 & 0x7f);
				}

				break;
			case 0x90:
				if ((cmd_offset+consumed+2)>cmd_len) {
					sprintf(text,"Voice: Note on (c=%d)",channel+1);
					if (usingrs) {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-(cmd_offset+1), text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_rsstatus, tvb, *rsoffset, 1, octet);
					} else {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-cmd_offset, text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
					}
					return 0;
				}
				octet2=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				octet3=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				/* note on with velocity=0 is interpreted as note off (to keep runningstate) */
				if (octet3)
				  sprintf(text,"Voice: Note on (c=%d,n=%d,v=%d)",channel+1,octet2&0x7f,octet3&0x7f);
				else
				  sprintf(text,"Voice: Note off* (c=%d,n=%d,v=%d)",channel+1,octet2&0x7f,octet3&0x7f);
				if (usingrs) {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rscommand, tvb, *rsoffset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rschannel, tvb, *rsoffset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_note, tvb, offset+cmd_offset, 1, octet2 & 0x7f);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_velocity, tvb, offset+cmd_offset+1, 1, octet3 & 0x7f);
				} else {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_command, tvb, offset+cmd_offset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_channel, tvb, offset+cmd_offset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_note, tvb, offset+cmd_offset+1, 1, octet2 & 0x7f);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_velocity, tvb, offset+cmd_offset+2, 1, octet3 & 0x7f);
				}
				break;
			case 0xa0:
				if ((cmd_offset+consumed+2)>cmd_len) {
					sprintf(text,"Voice: Key pressure (c=%d)",channel+1);
					if (usingrs) {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-(cmd_offset+1), text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_rsstatus, tvb, *rsoffset, 1, octet);
					} else {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-cmd_offset, text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
					}
					return 0;
				}
				octet2=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				octet3=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				sprintf(text,"Voice: Key pressure (c=%d,n=%d,p=%d)",channel+1,octet2&0x7f,octet3&0x7f);
				if (usingrs) {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rscommand, tvb, *rsoffset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rschannel, tvb, *rsoffset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_note, tvb, offset+cmd_offset, 1, octet2 & 0x7f);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_pressure, tvb, offset+cmd_offset+1, 1, octet3 & 0x7f);
				} else {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_command, tvb, offset+cmd_offset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_channel, tvb, offset+cmd_offset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_note, tvb, offset+cmd_offset+1, 1, octet2 & 0x7f);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_pressure, tvb, offset+cmd_offset+2, 1, octet3 & 0x7f);
				}
				break;
			case 0xb0:
				if ((cmd_offset+consumed+2)>cmd_len) {
					sprintf(text,"Voice: Controller change (c=%d)",channel+1);
					if (usingrs) {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-(cmd_offset+1), text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_rsstatus, tvb, *rsoffset, 1, octet);
					} else {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-cmd_offset, text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
					}
					return 0;
				}
				octet2=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				octet3=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;

				



				sprintf(text,"Voice: Controller change (c=%d,\"%s\",v=%d)",channel+1,val_to_str( octet2, rtp_midi_controller_values, "Unknown: %u" ),octet3&0x7f);
				if (usingrs) {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rscommand, tvb, *rsoffset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rschannel, tvb, *rsoffset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_controller, tvb, offset+cmd_offset, 1, octet2); 
					proto_tree_add_uint ( command_tree, hf_rtp_midi_controller_value, tvb, offset+cmd_offset+1, 1, octet3 & 0x7f);
				} else {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_command, tvb, offset+cmd_offset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_channel, tvb, offset+cmd_offset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_controller, tvb, offset+cmd_offset+1, 1, octet2);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_controller_value, tvb, offset+cmd_offset+2, 1, octet3 & 0x7f);
				}
				break;
			case 0xc0:
				if ((cmd_offset+consumed+1)>cmd_len) {
					sprintf(text,"Voice: Program change (c=%d)",channel+1);
					if (usingrs) {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-(cmd_offset+1), text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_rsstatus, tvb, *rsoffset, 1, octet);
					} else {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-cmd_offset, text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
					}
					return 0;
				}
				octet2=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				sprintf(text,"Voice: Program change (c=%d,p=%d)",channel+1,octet2&0x7f);
				if (usingrs) {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rscommand, tvb, *rsoffset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rschannel, tvb, *rsoffset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_program, tvb, offset+cmd_offset, 1, octet2 & 0x7f);
				} else {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_command, tvb, offset+cmd_offset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_channel, tvb, offset+cmd_offset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_program, tvb, offset+cmd_offset+1, 1, octet2 & 0x7f);
				}
				break;
			case 0xd0:
				if ((cmd_offset+consumed+1)>cmd_len) {
					sprintf(text,"Voice: Channel pressure (c=%d)",channel+1);
					if (usingrs) {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-(cmd_offset+1), text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_rsstatus, tvb, *rsoffset, 1, octet);
					} else {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-cmd_offset, text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
					}
					return 0;
				}
				octet2=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				sprintf(text,"Voice: Channel pressure (c=%d,p=%d)",channel+1,octet2&0x7f);
				if (usingrs) {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rscommand, tvb, *rsoffset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rschannel, tvb, *rsoffset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_channel_pressure, tvb, offset+cmd_offset, 1, octet2 & 0x7f);
				} else {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_command, tvb, offset+cmd_offset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_channel, tvb, offset+cmd_offset, 1, channel);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_channel_pressure, tvb, offset+cmd_offset+1, 1, octet2 & 0x7f);
				}
				break;
			case 0xe0:
				if ((cmd_offset+consumed+2)>cmd_len) {
					sprintf(text,"Voice: Pitch bend (c=%d)",channel+1);
					if (usingrs) {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-(cmd_offset+1), text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_rsstatus, tvb, *rsoffset, 1, octet);
					} else {
						command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, cmd_len-cmd_offset, text);
						command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
						proto_tree_add_uint ( command_tree, hf_rtp_midi_status, tvb, offset+cmd_offset, 1, octet);
					}
					return 0;
				}
				octet2=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;
				octet3=tvb_get_guint8(tvb,offset+cmd_offset+consumed);
				consumed++;

				wordvalue=octet3 & 0x7f;
				wordvalue<<=7;
				wordvalue|=octet2 & 0x7f;

				sprintf(text,"Voice: Pitch bend (c=%d,p=%d)",channel+1,wordvalue);
				if (usingrs) {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rscommand, tvb, *rsoffset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_rschannel, tvb, *rsoffset, 1, channel);
				    sprintf(helptext,"Pitchbend: 0x%02x",wordvalue);
				    command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset, 2, helptext);
				} else {
					command_item = proto_tree_add_text(tree, tvb, offset+cmd_offset, consumed, text);
					command_tree = proto_item_add_subtree( command_item, ett_rtp_midi_command);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_command, tvb, offset+cmd_offset, 1, cmd);
					proto_tree_add_uint ( command_tree, hf_rtp_midi_channel, tvb, offset+cmd_offset, 1, channel);
				    sprintf(helptext,"Pitchbend: 0x%02x",wordvalue);
				    command_item = proto_tree_add_text(command_tree, tvb, offset+cmd_offset+1, 2, helptext);
				}
				break;
		}
	} else {								/* should never get here... */
		return 0;
	}



	return consumed;
}

static void
dissect_rtp_midi( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree )
{
	proto_item *ti            = NULL;
	proto_item *command_item      = NULL;
	proto_item *journal_item      = NULL;
	proto_tree *rtp_midi_tree     = NULL;
	proto_tree *rtp_midi_commands_tree     = NULL;
	proto_tree *rtp_midi_journal_tree     = NULL;
	proto_tree *rtp_midi_sysjournal_tree     = NULL;
	proto_tree *rtp_midi_chanjournals_tree   = NULL;
	proto_tree *rtp_midi_chanjournal_tree     = NULL;
	unsigned int offset       = 0;
	struct _rtp_conversation_info *p_conv_data = NULL;

	guint8			flags;
	guint8			flags2;
	guint8			octet;
	guint8			totchan;
	guint16			check_seq_num;
	unsigned int    cmd_len;
	unsigned int	cmd_offset;
	unsigned int	cmd_count;
	unsigned int	consumed;
	guint8			runningstatus;
	unsigned int	rsoffset;
	char			text[160];
	char			info[170];
	char			help[100];
	unsigned int	i;
	guint16			sysjourlen;
	guint16			chanjourlen;
	guint16			systemflags;
	guint32			chanflags;
	guint8			channel;



	if ( check_col( pinfo->cinfo, COL_PROTOCOL ) )
	  {
	    col_set_str( pinfo->cinfo, COL_PROTOCOL, "RTP-MIDI" );
	  }
	if (check_col(pinfo->cinfo, COL_INFO))
		col_clear(pinfo->cinfo, COL_INFO);


	/* Get event fields */


/*	if (tree) { */

		/* get header flags */
		flags = tvb_get_guint8(tvb, offset );

		ti = proto_tree_add_item(tree,proto_rtp_midi,tvb,0,-1,FALSE);
	    rtp_midi_tree = proto_item_add_subtree( ti, ett_rtp_midi);

		proto_tree_add_boolean (rtp_midi_tree, hf_rtp_midi_bflag, tvb, offset, 1, flags);
		proto_tree_add_boolean (rtp_midi_tree, hf_rtp_midi_jflag, tvb, offset, 1, flags);
		proto_tree_add_boolean (rtp_midi_tree, hf_rtp_midi_zflag, tvb, offset, 1, flags);
		proto_tree_add_boolean (rtp_midi_tree, hf_rtp_midi_pflag, tvb, offset, 1, flags);

		cmd_len=flags & 0xf;

		/* see if we have small or large len-field */
		if ((flags & 0x80)==0x80)  {
			octet = tvb_get_guint8(tvb, offset+1 );
			cmd_len=(cmd_len << 8) | octet;
		    proto_tree_add_uint ( rtp_midi_tree, hf_rtp_midi_longlen, tvb, offset, 2, cmd_len);
			offset+=2;
		} else {
			proto_tree_add_uint ( rtp_midi_tree, hf_rtp_midi_shortlen, tvb, offset, 1, cmd_len);
			offset++;
		}

		/* if we have a command-section -> dissect it */
		if (cmd_len) {
	
		    command_item = proto_tree_add_text(rtp_midi_tree, tvb, offset, cmd_len, "Commands");
	        rtp_midi_commands_tree = proto_item_add_subtree( command_item, ett_rtp_midi_commands);

			/* if the reported command-length larger than data found in packet -> error */
			if ((offset+cmd_len)>tvb->length) {
				THROW(ReportedBoundsError);
				return;
			}

			cmd_offset=0;
		    cmd_count=0;
			/* RTP-MIDI-pdus always start with no running status */
			runningstatus=0;
		    while (cmd_offset<cmd_len) {

				
				/* for the first command we only have a delta-time if Z-Flag is set */
				if ((cmd_count!=0) || ((flags&0x20)==0x20)) {
					
					consumed=decodetime(tvb, pinfo, rtp_midi_commands_tree, offset, cmd_offset, cmd_len);
					if (0==consumed) {
						THROW(ReportedBoundsError);
						return;
					}
					cmd_offset+=consumed;

					/* we should have at least one midi-command coming after delta-time */
					if (cmd_offset>cmd_len) {
						THROW(ReportedBoundsError);
						return;
					}
				}

				/* dissect midi-command */
				consumed=decodemidi(tvb, pinfo, rtp_midi_commands_tree, offset, cmd_offset, cmd_len, &runningstatus, &rsoffset, text);
				if (0==consumed) {
					THROW(ReportedBoundsError);
					return;
				}
				cmd_offset+=consumed;
				
				cmd_count++;

				if ( check_col( pinfo->cinfo, COL_INFO) ) {
					if (1==cmd_count) {
						strcpy(info,text);
						col_add_fstr(pinfo->cinfo, COL_INFO,text);
					} else {
						if (2==cmd_count) {
							sprintf(text,"%s - one more command",info);
						} else {
							sprintf(text,"%s - %d more commands",info,cmd_count-1);
						}
						col_add_fstr(pinfo->cinfo, COL_INFO,text);
					}
				}
			}
			offset+=cmd_len;

		}

		/* if we have a journal-section -> dissect it */
		if ((flags & 0x40)==0x40) {
		  journal_item = proto_tree_add_text(rtp_midi_tree, tvb, offset, -1, "Journal");
	      rtp_midi_journal_tree = proto_item_add_subtree( journal_item, ett_rtp_midi_journal);
		  flags2 = tvb_get_guint8(tvb, offset );
		  proto_tree_add_boolean (rtp_midi_journal_tree, hf_rtp_midi_sflag, tvb, offset, 1, flags2);
		  proto_tree_add_boolean (rtp_midi_journal_tree, hf_rtp_midi_yflag, tvb, offset, 1, flags2);
		  proto_tree_add_boolean (rtp_midi_journal_tree, hf_rtp_midi_aflag, tvb, offset, 1, flags2);
		  proto_tree_add_boolean (rtp_midi_journal_tree, hf_rtp_midi_hflag, tvb, offset, 1, flags2);
		  totchan=flags2&0xf;
		  proto_tree_add_uint ( rtp_midi_journal_tree, hf_rtp_midi_totchan, tvb, offset, 1, totchan);
		  offset++;
		  check_seq_num=tvb_get_ntohs(tvb,offset);
		  proto_tree_add_uint ( rtp_midi_journal_tree, hf_rtp_midi_check_seq_num, tvb, offset, 2, check_seq_num);
		  offset+=2;

		  /* do we have system journal? */
		  if ((flags2 & 0x40)==0x40) {
				systemflags=tvb_get_ntohs(tvb,offset);
				sysjourlen=systemflags&0x3ff;
				ti = proto_tree_add_text(rtp_midi_journal_tree, tvb, offset, sysjourlen+2, "System-Journal");
				rtp_midi_sysjournal_tree = proto_item_add_subtree( ti, ett_rtp_midi_systemjournal);
				proto_tree_add_boolean (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_s, tvb, offset, 2, systemflags);
				proto_tree_add_boolean (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_d, tvb, offset, 2, systemflags);
				proto_tree_add_boolean (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_v, tvb, offset, 2, systemflags);
				proto_tree_add_boolean (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_q, tvb, offset, 2, systemflags);
				proto_tree_add_boolean (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_f, tvb, offset, 2, systemflags);
				proto_tree_add_boolean (rtp_midi_sysjournal_tree, hf_rtp_midi_sysjour_toc_x, tvb, offset, 2, systemflags);
				proto_tree_add_uint ( rtp_midi_journal_tree, hf_rtp_midi_totchan, tvb, offset, 2, systemflags);
				offset+=2;
				ti = proto_tree_add_text(rtp_midi_sysjournal_tree, tvb, offset, sysjourlen-2, "System Chapters");
		  }

		  /* do we have channel journal? */
		  if ((flags2 & 0x20)==0x20) {
			  	ti = proto_tree_add_text(rtp_midi_journal_tree, tvb, offset, -1, "Channel-Journal");
				rtp_midi_chanjournals_tree = proto_item_add_subtree( ti, ett_rtp_midi_channeljournals);
				/* iterate through all the channels specified in header */
				for (i=0;i<=totchan;i++) {
					chanflags=tvb_get_ntoh24(tvb,offset);
					channel=(chanflags>>19)&0xf;
					chanjourlen=(chanflags>>8)&0x3ff;
					sprintf(help,"Channel %d",channel+1);
			  		ti = proto_tree_add_text(rtp_midi_chanjournals_tree, tvb, offset, chanjourlen, help);
					rtp_midi_chanjournal_tree = proto_item_add_subtree( ti, ett_rtp_midi_channeljournal);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_sflag, tvb, offset, 3, chanflags);
				    proto_tree_add_uint ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_chan, tvb, offset, 3,  chanflags);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_hflag, tvb, offset, 3, chanflags);
				    proto_tree_add_uint ( rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_len, tvb, offset, 3,  chanflags);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_p, tvb, offset, 3, chanflags);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_c, tvb, offset, 3, chanflags);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_m, tvb, offset, 3, chanflags);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_w, tvb, offset, 3, chanflags);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_n, tvb, offset, 3, chanflags);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_e, tvb, offset, 3, chanflags);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_t, tvb, offset, 3, chanflags);
					proto_tree_add_boolean (rtp_midi_chanjournal_tree, hf_rtp_midi_chanjour_toc_a, tvb, offset, 3, chanflags);
					ti = proto_tree_add_text(rtp_midi_chanjournal_tree, tvb, offset+3, chanjourlen-3, "Channel Chapters");
					offset+=chanjourlen;
				}
		  }

		}
/*	} /* if (tree) */

}




void
proto_register_rtp_midi(void)
{

	module_t *rtp_midi_module;

	static hf_register_info hf[] =
	{
		{
			&hf_rtp_midi_bflag,
			{
				"B-Flag",
				"rtpmidi.b_flag",
				FT_BOOLEAN,
				8,
				NULL,
				0x80,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_jflag,
			{
				"J-Flag",
				"rtpmidi.j_flag",
				FT_BOOLEAN,
				8,
				NULL,
				0x40,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_zflag,
			{
				"Z-Flag",
				"rtpmidi.z_flag",
				FT_BOOLEAN,
				8,
				NULL,
				0x20,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_pflag,
			{
				"P-Flag",
				"rtpmidi.p_flag",
				FT_BOOLEAN,
				8,
				NULL,
				0x10,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_shortlen,
			{
				"Command length (short)",
				"rtpmidi.cmd_length_short",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x0F,
				"", HFILL
			}
		},

		{
			&hf_rtp_midi_longlen,
			{
				"Command length (long)",
				"rtpmidi.cmd_length_long",
				FT_UINT16,
				BASE_DEC,
				NULL,
				0x0FFF,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_sflag,
			{
				"S-Flag",
				"rtpmidi.s_flag",
				FT_BOOLEAN,
				8,
				NULL,
				0x80,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_yflag,
			{
				"Y-Flag",
				"rtpmidi.y_flag",
				FT_BOOLEAN,
				8,
				NULL,
				0x40,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_aflag,
			{
				"A-Flag",
				"rtpmidi.a_flag",
				FT_BOOLEAN,
				8,
				NULL,
				0x20,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_hflag,
			{
				"H-Flag",
				"rtpmidi.h_flag",
				FT_BOOLEAN,
				8,
				NULL,
				0x10,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_totchan,
			{
				"Total channels",
				"rtpmidi.total_channels",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x0F,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_check_seq_num,
			{
				"Checkpoint Packet Seqnum",
				"rtpmidi.check_Seq_num",
				FT_UINT16,
				BASE_DEC,
				NULL,
				0x00,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_deltatime1,
			{
				"Delta Time (one octet)",
				"rtpmidi.deltatime_1",
				FT_UINT8,
				BASE_HEX,
				NULL,
				0,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_deltatime2,
			{
				"Delta Time (two octets)",
				"rtpmidi.deltatime_2",
				FT_UINT16,
				BASE_HEX,
				NULL,
				0,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_deltatime3,
			{
				"Delta Time (three octets)",
				"rtpmidi.deltatime_3",
				FT_UINT24,
				BASE_HEX,
				NULL,
				0,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_deltatime4,
			{
				"Delta Time (four octets)",
				"rtpmidi.deltatime_4",
				FT_UINT32,
				BASE_HEX,
				NULL,
				0,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_status,
			{
				"Status",
				"rtpmidi.status",
				FT_UINT8,
				BASE_HEX,
				NULL,
				0,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_rsstatus,
			{
				"RS Status",
				"rtpmidi.rsstatus",
				FT_UINT8,
				BASE_HEX,
				NULL,
				0,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_command,
			{
				"Command",
				"rtpmidi.command",
				FT_UINT8,
				BASE_HEX,
				NULL,
				0xf0,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_rscommand,
			{
				"RS Command",
				"rtpmidi.rscommand",
				FT_UINT8,
				BASE_HEX,
				NULL,
				0xf0,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_channel,
			{
				"Channel",
				"rtpmidi.channel",
				FT_UINT8,
				BASE_HEX,
				NULL,
				0x0f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_rschannel,
			{
				"RS Channel",
				"rtpmidi.rschannel",
				FT_UINT8,
				BASE_HEX,
				NULL,
				0x0f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_note,
			{
				"Note",
				"rtpmidi.note",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x7f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_velocity,
			{
				"Velocity",
				"rtpmidi.velocity",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x7f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_pressure,
			{
				"Pressure",
				"rtpmidi.pressure",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x7f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_controller,
			{
				"Controller",
				"rtpevent.controller",
				FT_UINT8,
				BASE_DEC,
				VALS(rtp_midi_controller_values),
				0x7f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_controller_value,
			{
				"Value",
				"rtpmidi.controller_value",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x7f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_program,
			{
				"Program",
				"rtpmidi.program",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x7f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_channel_pressure,
			{
				"Pressure",
				"rtpmidi.channel_pressure",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x7f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_manu_short,
			{
				"Manufacturer (short)",
				"rtpmidi.manufacturer_short",
				FT_UINT8,
				BASE_HEX,
				VALS(rtp_midi_manu_short_values),
				0x7f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_manu_long,
			{
				"Manufacturer (long)",
				"rtpmidi.manufacturer_long",
				FT_UINT16,
				BASE_HEX,
				VALS(rtp_midi_manu_long_values),
				0x7f7f,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_sysjour_toc_s,
			{
				"System-Journal TOC-S-Flag",
				"rtpmidi.sysjour_toc_s",
				FT_BOOLEAN,
				16,
				NULL,
				0x8000,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_sysjour_toc_d,
			{
				"System-Journal TOC-D-Flag",
				"rtpmidi.sysjour_toc_d",
				FT_BOOLEAN,
				16,
				NULL,
				0x4000,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_sysjour_toc_v,
			{
				"System-Journal TOC-V-Flag",
				"rtpmidi.sysjour_toc_v",
				FT_BOOLEAN,
				16,
				NULL,
				0x2000,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_sysjour_toc_q,
			{
				"System-Journal TOC-Q-Flag",
				"rtpmidi.sysjour_toc_q",
				FT_BOOLEAN,
				16,
				NULL,
				0x1000,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_sysjour_toc_f,
			{
				"System-Journal TOC-F-Flag",
				"rtpmidi.sysjour_toc_f",
				FT_BOOLEAN,
				16,
				NULL,
				0x0800,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_sysjour_toc_x,
			{
				"System-Journal TOC-X-Flag",
				"rtpmidi.sysjour_toc_x",
				FT_BOOLEAN,
				16,
				NULL,
				0x0400,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_sysjour_len,
			{
				"System-Journal-Length",
				"rtpmidi.cmd_sysjour_len",
				FT_UINT16,
				BASE_DEC,
				NULL,
				0x03FF,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_sflag,
			{
				"Channel-Journal S-Flag",
				"rtpmidi.chanjour_s",
				FT_BOOLEAN,
				24,
				NULL,
				0x800000,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_chan,
			{
				"Channel",
				"rtpmidi.chanjour_channel",
				FT_UINT24,
				BASE_HEX,
				NULL,
				0x780000,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_hflag,
			{
				"Channel-Journal H-Flag",
				"rtpmidi.chanjour_h",
				FT_BOOLEAN,
				24,
				NULL,
				0x040000,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_len,
			{
				"Channel-Journal-Length",
				"rtpmidi.cmd_chanjour_len",
				FT_UINT24,
				BASE_DEC,
				NULL,
				0x03FF00,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_toc_p,
			{
				"Channel-Journal TOC-P-Flag",
				"rtpmidi.chanjour_toc_p",
				FT_BOOLEAN,
				24,
				NULL,
				0x80,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_toc_c,
			{
				"Channel-Journal TOC-C-Flag",
				"rtpmidi.chanjour_toc_c",
				FT_BOOLEAN,
				24,
				NULL,
				0x40,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_toc_m,
			{
				"Channel-Journal TOC-M-Flag",
				"rtpmidi.chanjour_toc_m",
				FT_BOOLEAN,
				24,
				NULL,
				0x20,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_toc_w,
			{
				"Channel-Journal TOC-W-Flag",
				"rtpmidi.chanjour_toc_w",
				FT_BOOLEAN,
				24,
				NULL,
				0x10,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_toc_n,
			{
				"Channel-Journal TOC-N-Flag",
				"rtpmidi.chanjour_toc_n",
				FT_BOOLEAN,
				24,
				NULL,
				0x08,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_toc_e,
			{
				"Channel-Journal TOC-E-Flag",
				"rtpmidi.chanjour_toc_e",
				FT_BOOLEAN,
				24,
				NULL,
				0x04,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_toc_t,
			{
				"Channel-Journal TOC-T-Flag",
				"rtpmidi.chanjour_toc_t",
				FT_BOOLEAN,
				24,
				NULL,
				0x02,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_chanjour_toc_a,
			{
				"Channel-Journal TOC-A-Flag",
				"rtpmidi.chanjour_toc_a",
				FT_BOOLEAN,
				24,
				NULL,
				0x01,
				"", HFILL
			}
		},
		{
			&hf_rtp_midi_sysex_chan,
			{
				"Sysex-Channel",
				"rtpmidi.sysex_chan",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0,
				"", HFILL
			}
		},

	};


	static gint *ett[] =
	{
		&ett_rtp_midi,
		&ett_rtp_midi_commands,
		&ett_rtp_midi_journal,
		&ett_rtp_midi_command,
		&ett_rtp_midi_systemjournal,
		&ett_rtp_midi_channeljournals,
		&ett_rtp_midi_systemchapters,
		&ett_rtp_midi_channeljournal,
		&ett_rtp_midi_channelchapters,
		&ett_rtp_midi_sysex_data,
	};





	proto_rtp_midi = proto_register_protocol("RTP-MIDI", "RTP-MIDI", "rtpmidi");
	proto_register_field_array(proto_rtp_midi, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));


    /* Register preferences */
    rtp_midi_module = prefs_register_protocol (proto_rtp_midi, proto_reg_handoff_rtp_midi);
    prefs_register_uint_preference (rtp_midi_module,
                                    "midi_payload_type_value", "Payload Type for RFC 4695 RTP-MIDI",
                                    "This is the value of the Payload Type field "
                                    "that specifies RTP-MIDI", 10,
                                    &rtp_midi_payload_type_value);
	register_dissector("rtpmidi", dissect_rtp_midi, proto_rtp_midi);
}



void
proto_reg_handoff_rtp_midi(void)
{
	static dissector_handle_t rtp_midi_handle;
	static int rtp_midi_prefs_initialized = FALSE;


  	if (!rtp_midi_prefs_initialized) {
		rtp_midi_handle = create_dissector_handle(dissect_rtp_midi, proto_rtp_midi);
		rtp_midi_prefs_initialized = TRUE;
	}
	else {
		dissector_delete("rtp.pt", saved_payload_type_value, rtp_midi_handle);
	}

	saved_payload_type_value = rtp_midi_payload_type_value;
	/* rtp_event_payload_type_value is set from preferences */

    dissector_add("rtp.pt", saved_payload_type_value, rtp_midi_handle);
	dissector_add_string("rtp_dyn_payload_type", "midi", rtp_midi_handle);
}
